/*
** Libdax
**
** Copyright © 2016 Oracle corp.  All rights reserved.
** The Universal Permissive License (UPL), Version 1.0
**
** Subject to the condition set forth below, permission is hereby granted to any person obtaining a copy of this
** software, associated documentation and/or data (collectively the "Software"), free of charge and under any and
** all copyright rights in the Software, and any and all patent rights owned or freely licensable by each licensor
** hereunder covering either (i) the unmodified Software as contributed to or provided by such licensor, or
** (ii) the Larger Works (as defined below), to deal in both
**
** (a) the Software, and
** (b) any piece of software and/or hardware listed in the lrgrwrks.txt file if one is included with the Software
** (each a “Larger Work” to which the Software is contributed by such licensors),
**
** without restriction, including without limitation the rights to copy, create derivative works of, display,
** perform, and distribute the Software and make, use, sell, offer for sale, import, export, have made, and have
** sold the Software and the Larger Work(s), and to sublicense the foregoing rights on either these or other terms.
**
** This license is subject to the following condition:
** The above copyright notice and either this complete permission notice or at a minimum a reference to the UPL must
** be included in all copies or substantial portions of the Software.
**
** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
** THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
** AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
** CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
** IN THE SOFTWARE.
*/

#ifndef	_SYS_CCB_H
#define	_SYS_CCB_H

#ifdef __cplusplus
extern "C" {
#endif

/* CCB address types */
#define	CCB_AT_VA		3	/* virtual address */
#ifdef _KERNEL
#define	CCB_AT_RA		2	/* only kernel can use real address */
#endif /* _KERNEL */

#define	CCB_AT_COMPL_MASK	0x3
#define	CCB_AT_SRC0_MASK	0x7
#define	CCB_AT_SRC1_MASK	0x7
#define	CCB_AT_DST_MASK		0x7
#define	CCB_AT_TBL_MASK		0x3

#define	CCB_AT_COMPL_SHIFT	32
#define	CCB_AT_SRC0_SHIFT	34

/* CCB header sync flags */
#define	CCB_SYNC_SERIAL		0x1
#define	CCB_SYNC_COND		0x2
#define	CCB_SYNC_LONGCCB	0x4

#define	CCB_SYNC_FLG_SHIFT	24
#define	CCB_HDR_SHIFT		32

#define	CCB_DW1_INTR_SHIFT	59

#ifndef _ASM

#ifndef __linux
#include <sys/inttypes.h>
#endif

/* CCB header version */
#define	CCB_VERSION		0x0

/* CCB L3 output allocation */
#define	CCB_OUTPUT_ALLOC_NONE	0	/* do not allocate in L3 */
#define	CCB_OUTPUT_ALLOC_HARD	1	/* allocate in L3 of running cpu */
#define	CCB_OUTPUT_ALLOC_SOFT	2	/* allocate to whichever L3 owns */
					/* line, else L3 of running cpu */

#define	CCB_VA_MASK	((1ULL << 54) - 1)
#define	CCB_LOCAL_ADDR_SHIFT	6
#define	CCB_LOCAL_ADDR(x, mask)	(((x) & mask) >> CCB_LOCAL_ADDR_SHIFT)

#define	CCB_DWORD_CTL			0
#define	CCB_DWORD_COMPL			1
#define	MSG_DWORD_SRC		2
#define	MSG_DWORD_DST		6
#define	QUERY_DWORD_INPUT		2
#define	QUERY_DWORD_DAC			3
#define	QUERY_DWORD_SEC_INPUT		4
#define	QUERY_DWORD_OPERANDS_UL_MSW	5
#define	QUERY_DWORD_OUTPUT		6
#define	QUERY_DWORD_TBL			7
#define	QUERY_DWORD_OPERANDS_UL_MSW2	8
#define	QUERY_DWORD_OPERANDS_UL_MSW3	9
#define	QUERY_DWORD_OPERANDS_UL_MSW4	10

#define	BIT_MASK64(_hi, _lo)	(((uint64_t)((~(uint64_t)0)>>(63-(_hi)))) & \
	((uint64_t)((~(uint64_t)0)<<(_lo))))

#define	CCB_GET(s, dword)	(((dword) & CCB_##s##_MASK) >> CCB_##s##_SHIFT)

#define	CCB_SET(s, val, dword)				\
	((dword) = ((dword) & ~CCB_##s##_MASK) |	\
	((((val) << CCB_##s##_SHIFT)) & CCB_##s##_MASK))

#define	CCB_COMPL_VA_MASK		BIT_MASK64(53, 6)
#define	CCB_COMPL_VA_SHIFT		0

#define	CCB_COMPL_PA_MASK		BIT_MASK64(55, 6)
#define	CCB_COMPL_PA_SHIFT		0

#define	CCB_QUERY_TBL_VA_MASK		CCB_COMPL_VA_MASK
#define	CCB_QUERY_TBL_VA_SHIFT		CCB_COMPL_VA_SHIFT

#define	CCB_QUERY_TBL_PA_MASK		CCB_COMPL_PA_MASK
#define	CCB_QUERY_TBL_PA_SHIFT		CCB_COMPL_PA_SHIFT

#define	CCB_QUERY_OUTPUT_VA_MASK	CCB_COMPL_VA_MASK
#define	CCB_QUERY_OUTPUT_VA_SHIFT	CCB_COMPL_VA_SHIFT

#define	CCB_QUERY_OUTPUT_PA_MASK	CCB_COMPL_PA_MASK
#define	CCB_QUERY_OUTPUT_PA_SHIFT	CCB_COMPL_PA_SHIFT

#define	CCB_QUERY_INPUT_VA_MASK		BIT_MASK64(53, 0)
#define	CCB_QUERY_INPUT_VA_SHIFT	0

#define	CCB_QUERY_INPUT_PA_MASK		BIT_MASK64(55, 0)
#define	CCB_QUERY_INPUT_PA_SHIFT	0

#define	CCB_QUERY_SEC_INPUT_VA_MASK	CCB_QUERY_INPUT_VA_MASK
#define	CCB_QUERY_SEC_INPUT_VA_SHIFT	CCB_QUERY_INPUT_VA_SHIFT

#define	CCB_QUERY_SEC_INPUT_PA_MASK	CCB_QUERY_INPUT_PA_MASK
#define	CCB_QUERY_SEC_INPUT_PA_SHIFT	CCB_QUERY_INPUT_PA_SHIFT

#define	CCB_COMPL_VA(dword)		CCB_GET(COMPL_VA, (dword))

#define	CCB_QUERY_INPUT_VA(dword)	CCB_GET(QUERY_INPUT_VA, (dword))
#define	CCB_QUERY_SEC_INPUT_VA(dword)	CCB_GET(QUERY_SEC_INPUT_VA, (dword))
#define	CCB_QUERY_OUTPUT_VA(dword)	CCB_GET(QUERY_OUTPUT_VA, (dword))
#define	CCB_QUERY_TBL_VA(dword)		CCB_GET(QUERY_TBL_VA, (dword))

#define	CCB_SET_COMPL_PA(pa, dword)	CCB_SET(COMPL_PA, pa, dword)

#define	CCB_SET_QUERY_INPUT_PA(pa, dw)	CCB_SET(QUERY_INPUT_PA, (pa), (dw))
#define	CCB_SET_QUERY_SEC_INPUT_PA(pa, dw)	\
	CCB_SET(QUERY_SEC_INPUT_PA, (pa), (dw))
#define	CCB_SET_QUERY_OUTPUT_PA(pa, dw)	CCB_SET(QUERY_OUTPUT_PA, (pa), (dw))
#define	CCB_SET_QUERY_TBL_PA(pa, dw)	CCB_SET(QUERY_TBL_PA, (pa), (dw))

/*
 * Message CCB opcodes
 */
#define	CCB_MSG_OPCODE_SYNC_NOP		0x80
#define	CCB_MSG_OPCODE_COPY		0x81
#define	CCB_MSG_OPCODE_FILL		0x83
#define	CCB_MSG_OPCODE_STORE		0x84
#define	CCB_MSG_OPCODE_CAM		0x85
#define	CCB_MSG_OPCODE_CAS		0x86
#define	CCB_MSG_OPCODE_CAS_FETCH	0x87
#define	CCB_MSG_OPCODE_ATOMIC_ADD	0x88
#define	CCB_MSG_OPCODE_ATOMIC_OR	0x89
#define	CCB_MSG_OPCODE_ATOMIC_PUSH	0x8A
#define	CCB_MSG_OPCODE_COPY_IMM		0xC1

/* CCB_OPCODE_SYNC_NOP extended opcodes */
#define	CCB_OPCODE_EXT_NOP		0x0
#define	CCB_OPCODE_EXT_SYNC		0x1

/*
 * Query CCB opcodes
 */
#define	CCB_QUERY_OPCODE_SYNC_NOP	0x0
#define	CCB_QUERY_OPCODE_EXTRACT	0x1
#define	CCB_QUERY_OPCODE_SCAN_VALUE	0x2
#define	CCB_QUERY_OPCODE_SCAN_RANGE	0x3
#define	CCB_QUERY_OPCODE_TRANSLATE	0x4
#define	CCB_QUERY_OPCODE_SELECT		0x5
#define	CCB_QUERY_OPCODE_INV_SCAN_VALUE	0x12
#define	CCB_QUERY_OPCODE_INV_SCAN_RANGE	0x13
#define	CCB_QUERY_OPCODE_INV_TRANSLATE	0x14

/* Query primary input formats */
#define	CCB_QUERY_IFMT_FIX_BYTE		0	/* to 16 bytes */
#define	CCB_QUERY_IFMT_FIX_BIT		1	/* to 15 bits */
#define	CCB_QUERY_IFMT_VAR_BYTE		2	/* separate length stream */
#define	CCB_QUERY_IFMT_FIX_BYTE_RLE	4	/* to 16 bytes + RL stream */
#define	CCB_QUERY_IFMT_FIX_BIT_RLE	5	/* to 15 bits + RL stream */
#define	CCB_QUERY_IFMT_FIX_BYTE_HUFF	8	/* to 16 bytes */
#define	CCB_QUERY_IFMT_FIX_BIT_HUFF	9	/* to 15 bits */
#define	CCB_QUERY_IFMT_VAR_BYTE_HUFF	0xA	/* separate length stream */
#define	CCB_QUERY_IFMT_FIX_BYTE_RLE_HUFF 0xC	/* to 16 bytes + RL stream */
#define	CCB_QUERY_IFMT_FIX_BIT_RLE_HUFF	0xD	/* to 15 bits + RL stream */

/* Query secondary input size */
#define	CCB_QUERY_SZ_ONEBIT		0
#define	CCB_QUERY_SZ_TWOBIT		1
#define	CCB_QUERY_SZ_FOURBIT		2
#define	CCB_QUERY_SZ_EIGHTBIT		3

/* Query secondary input encoding */
#define	CCB_QUERY_SIE_LESS_ONE		0
#define	CCB_QUERY_SIE_ACTUAL		1

/* Query output formats */
#define	CCB_QUERY_OFMT_BYTE_ALIGN	0
#define	CCB_QUERY_OFMT_16B		1
#define	CCB_QUERY_OFMT_BIT_VEC		2
#define	CCB_QUERY_OFMT_ONE_IDX		3

/* Query operand size constants */
#define	CCB_QUERY_OPERAND_DISABLE	31

/* Query Data Access Control input length format */
#define	CCB_QUERY_ILF_SYMBOL		0
#define	CCB_QUERY_ILF_BYTE		1
#define	CCB_QUERY_ILF_BIT		2

/* Completion area cmd_status */
#define	CCB_CMD_STAT_NOT_COMPLETED	0
#define	CCB_CMD_STAT_COMPLETED		1
#define	CCB_CMD_STAT_FAILED		2
#define	CCB_CMD_STAT_KILLED		3
#define	CCB_CMD_STAT_NOT_RUN		4

/* Completion area err_mask of user visible errors */
#define	CCB_CMD_ERR_BOF			0x1	/* buffer overflow */
#define	CCB_CMD_ERR_DECODE		0x2	/* CCB decode error */
#define	CCB_CMD_ERR_POF			0x3	/* page overflow */
#define	CCB_CMD_ERR_KILL		0x7	/* command was killed */
#define	CCB_CMD_ERR_ADI			0x9	/* ADI error */
#define	CCB_CMD_ERR_DATA_FMT		0xA	/* data format error */
#define	CCB_CMD_ERR_OTHER_NO_RETRY	0xE	/* Other error, do not retry */
#define	CCB_CMD_ERR_OTHER_RETRY		0xF	/* Other error, retry */
#define	CCB_CMD_ERR_PARTIAL_SYMBOL	0x80	/* QP partial symbol warning */

typedef struct ccb_hdr {
	uint32_t	ccb_ver:4;	/* must be set to 0 for M7 HW */
	uint32_t	sync_flags:4;
	uint32_t	opcode:8;
	uint32_t	rsvd:3;
	uint32_t	at_tbl:2;	/* RA(kernel)/VA */
	uint32_t	at_dst:3;	/* RA(kernel)/VA */
	uint32_t	at_src1:3;	/* RA(kernel)/VA */
	uint32_t	at_src0:3;	/* RA(kernel)/VA */
#ifdef _KERNEL
	uint32_t	at_cmpl:2;	/* RA(kernel)/VA */
#else
	uint32_t	rsvd2:2;	/* only kernel can specify at_cmpl */
#endif /* _KERNEL */
} ccb_hdr_t;

typedef struct ccb {
	uint64_t		dwords[8];
} ccb_t;

typedef struct ccb_completion_area {
	volatile uint8_t cmd_status;	/* user may mwait on this address */
	uint8_t		err_mask;	/* user visible error notification */
	uint8_t		rsvd[2];	/* reserved */
	uint32_t	err_vals;	/* for QP partial symbol warning */
	uint32_t	output_sz;	/* query ops only */
	uint32_t	rsvd2;		/* reserved */
	uint64_t	run_time;	/* run time in OCND2 cycles */
	uint64_t	run_stats;	/* nothing reported in version 1.0 */
	uint32_t	n_processed;	/* query ops only */
	uint32_t	rsvd3[5];	/* reserved */
	uint64_t	command_rv;	/* command return value */
	uint64_t	rsvd4[8];	/* reserved */
} ccb_completion_area_t;


typedef struct ccb_addr {
	uint64_t	adp:4;
	uint64_t	rsvd:4;
	uint64_t	addr:50;	/* [55:6] of 64B aligned address */
					/* if VA, [55:54] must be 0 */
	uint64_t	rsvd2:6;
} ccb_addr_t;

typedef struct ccb_byte_addr {
	uint64_t	adp:4;
	uint64_t	rsvd:4;
	uint64_t	addr:56;	/* [55:0] of byte aligned address */
					/* if VA, [55:54] must be 0 */
} ccb_byte_addr_t;

typedef struct ccb_st_dst_addr {
	uint64_t	adp:4;
	uint64_t	rsvd:4;
	uint64_t	addr:50;	/* [55:6] of byte aligned address */
					/* if VA, [55:54] must be 0 */
	uint64_t	rsvd2:6;	/* [5:0] is in control word */
} ccb_st_dst_addr_t;

typedef struct ccb_tbl_addr {
	uint64_t	adp:4;
	uint64_t	rsvd:4;
	uint64_t	addr:50;	/* [55:6] of 64B aligned address */
					/* if VA, [55:54] must be 0 */
	uint64_t	rsvd2:4;
	uint64_t	vers:2;		/* version number */
} ccb_tbl_addr_t;

typedef struct ccb_cmpl_addr {
	uint64_t	adp:4;
	uint64_t	intr:1;		/* if 1, send interrupt on completion */
#ifdef _KERNEL
	uint64_t	rsvd:3;
	uint64_t	addr:50;	/* [55:6] of 64B aligned address */
					/* if VA, [55:54] must be 0 */
	uint64_t	rsvd2:6;
#else
	uint64_t	rsvd:59;	/* Only kernel can specify completion */
					/* address in CCB.  User must use */
					/* offset to mmapped kernel memory. */
#endif /* _KERNEL */
} ccb_cmpl_addr_t;

/*
 * Message CCB definitions
 */

typedef struct ccb_msg_ctl {
	ccb_hdr_t	hdr;
	uint32_t	src_adp:4;	/* if source is RKEY, else 0 */
	uint32_t	dst_adp:4;	/* if dest is RKEY, else 0 */
	uint32_t	size:24;	/* number of 64B cache lines, less 1 */
} ccb_msg_ctl_t;

/*
 * CCB_MSG_OPCODE_COPY
 */
typedef struct ccb_copy {
	ccb_msg_ctl_t	ctl;
	ccb_cmpl_addr_t completion;
	union {
		uint64_t	rkey_hi;
		ccb_addr_t	local;
	} src0;
	union {
		uint64_t	rkey_lo;
		uint64_t	rsvd;
	} src1;
	uint64_t	rsvd[2];
	union {
		uint64_t	rkey_hi;
		ccb_addr_t	local;
	} dst0;
	union {
		uint64_t	rkey_lo;
		struct {
			uint32_t	output_alloc:2;
			uint32_t	rsvd:30;
			uint32_t	rsvd2;
		} local;
	} dst1;
} ccb_copy_t;

/*
 * Copy immediate data payload for CCB_MSG_OPCODE_COPY_IMM.
 */
typedef struct ccb_imm_data {
	uint64_t	data[8];
} ccb_imm_data_t;

typedef struct ccb_sync_nop_ctl {
	ccb_hdr_t	hdr;
	uint32_t	ext_op:1;	/* extended op flag */
	uint32_t	rsvd:31;
} ccb_sync_nop_ctl_t;

/*
 * CCB_MSG_OPCODE_SYNC_NOP
 * CCB_QUERY_OPCODE_SYNC_NOP
 */
typedef struct ccb_sync_nop {
	ccb_sync_nop_ctl_t	ctl;
	ccb_cmpl_addr_t		completion;
	uint64_t		rsvd[6];
} ccb_sync_nop_t;

typedef struct ccb_fill_ctl {
	ccb_hdr_t	hdr;
	uint32_t	rsvd:4;
	uint32_t	dst_adp:4;	/* if dest is RKEY, else 0 */
	uint32_t	size:24;	/* number of 64B cache lines, less 1 */
} ccb_fill_ctl_t;

/*
 * CCB_MSG_OPCODE_FILL
 */
typedef struct ccb_fill {
	ccb_fill_ctl_t		ctl;
	ccb_cmpl_addr_t		completion;
	uint64_t		imm_op;		/* fill value */
	uint64_t		rsvd[3];
	union {
		uint64_t	rkey_hi;
		ccb_addr_t	local;
	} dst0;
	union {
		uint64_t	rkey_lo;
		struct {
			uint32_t	output_alloc:2;
			uint32_t	rsvd:30;
			uint32_t	rsvd2;
		} local;
	} dst1;
} ccb_fill_t;

typedef struct ccb_store_ctl {
	ccb_hdr_t	hdr;
	uint32_t	rsvd:4;
	uint32_t	dst_adp:4;	/* if dest is RKEY, else 0 */
	uint32_t	rsvd2:10;
	uint32_t	dst_offset:6;	/* [5:0] of destination address */
					/* byte offset within cache line */
	uint32_t	rsvd3:6;
	uint32_t	size:2;		/* CCB_CW_SZ_[1|2|4|8]_BYTE */
} ccb_store_ctl_t;

/*
 * CCB_MSG_OPCODE_STORE
 */
typedef struct ccb_store {
	ccb_store_ctl_t		ctl;
	ccb_cmpl_addr_t		completion;
	uint64_t		imm_op;		/* fill value */
	uint64_t		rsvd[3];
	union {
		uint64_t	rkey_hi;
		ccb_st_dst_addr_t local;
	} dst0;
	union {
		uint64_t	rkey_lo;
		struct {
			uint32_t	output_alloc:2;
			uint32_t	rsvd:30;
			uint32_t	rsvd2;
		} local;
	} dst1;
} ccb_store_t;

/*
 * CCB_MSG_OPCODE_CAS
 * CCB_MSG_OPCODE_CAS_FETCH
 */
typedef struct ccb_cas {
	ccb_store_ctl_t		ctl;
	ccb_cmpl_addr_t		completion;
	uint64_t		swp_val;
	uint64_t		rsvd;
	uint64_t		cmp_val;
	uint64_t		rsvd2;
	union {
		uint64_t	rkey_hi;
		ccb_st_dst_addr_t local;
	} dst0;
	union {
		uint64_t	rkey_lo;
		struct {
			uint32_t	output_alloc:2;
			uint32_t	rsvd:30;
			uint32_t	rsvd2;
		} local;
	} dst1;
} ccb_cas_t;

/*
 * CCB_MSG_OPCODE_CAM
 * CCB_MSG_OPCODE_ATOMIC_ADD
 * CCB_MSG_OPCODE_ATOMIC_OR
 */
typedef struct ccb_cam_at {
	ccb_store_ctl_t		ctl;
	ccb_cmpl_addr_t		completion;
	uint64_t		imm_op;
	uint64_t		rsvd[3];
	union {
		uint64_t	rkey_hi;
		ccb_st_dst_addr_t local;
	} dst0;
	union {
		uint64_t	rkey_lo;
		struct {
			uint32_t	output_alloc:2;
			uint32_t	rsvd:30;
			uint32_t	rsvd2;
		} local;
	} dst1;
} ccb_cam_at_t;

typedef struct ccb_at_msg_ctl {
	ccb_hdr_t	hdr;
	uint32_t	src_adp:4;
	uint32_t	hmq_adp:4;
	uint32_t	rsvd:20;
	uint32_t	msg_size:4;	/* LOG2 number of 64B cache lines */
} ccb_at_msg_ctl_t;

/*
 * CCB_MSG_OPCODE_ATOMIC_PUSH
 */
typedef struct ccb_at_msg_push {
	ccb_at_msg_ctl_t	ctl;
	ccb_cmpl_addr_t		completion;
	ccb_addr_t		src;
	uint64_t		rsvd[3];
	union {
		uint64_t	rkey_hi;
		ccb_addr_t	local;
	} dst0;
	union {
		uint64_t	rkey_lo;
		uint64_t	rsvd;
	} dst1;
} ccb_at_msg_push_t;

/*
 * Query CCB definitions
 */

typedef struct ccb_extract_ctl {
	ccb_hdr_t	hdr;
	uint32_t	src0_fmt:4;
	uint32_t	src0_sz:5;
	uint32_t	src0_off:3;
	uint32_t	src1_enc:1;
	uint32_t	src1_off:3;
	uint32_t	src1_sz:2;
	uint32_t	output_fmt:2;
	uint32_t	output_sz:2;
	uint32_t	pad_dir:1;
	uint32_t	rsvd:9;
} ccb_extract_ctl_t;

typedef struct ccb_data_acc_ctl {
	uint64_t	flow_ctl:2;
	uint64_t	pipeline_targ:2;
	uint64_t	output_buf_sz:20;
	uint64_t	rsvd:8;
	uint64_t	output_alloc:2;
	uint64_t	rsvd2:4;
	uint64_t	input_len_fmt:2;
	uint64_t	input_cnt:24;
} ccb_data_acc_ctl_t;

/*
 * CCB_QUERY_OPCODE_EXTRACT
 */
typedef struct ccb_extract {
	ccb_extract_ctl_t	control;
	ccb_cmpl_addr_t		completion;
	ccb_byte_addr_t		src0;
	ccb_data_acc_ctl_t	data_acc_ctl;
	ccb_byte_addr_t		src1;
	uint64_t		rsvd;
	ccb_addr_t		output;
	ccb_tbl_addr_t		tbl;
} ccb_extract_t;

typedef struct ccb_scan_bound {
	uint32_t	upper;
	uint32_t	lower;
} ccb_scan_bound_t;

/*
 * CCB_QUERY_OPCODE_SCAN_VALUE
 * CCB_QUERY_OPCODE_SCAN_RANGE
 */
typedef struct ccb_scan {
	ccb_extract_ctl_t	control;
	ccb_cmpl_addr_t		completion;
	ccb_byte_addr_t		src0;
	ccb_data_acc_ctl_t	data_acc_ctl;
	ccb_byte_addr_t		src1;
	ccb_scan_bound_t	bound_msw;
	ccb_addr_t		output;
	ccb_tbl_addr_t		tbl;
} ccb_scan_t;

/*
 * Scan Value/Range words 8-15 required when L or U operand size > 4 bytes.
 */
typedef struct ccb_scan_ext {
	ccb_scan_bound_t	bound_msw2;
	ccb_scan_bound_t	bound_msw3;
	ccb_scan_bound_t	bound_msw4;
	uint64_t		rsvd[5];
} ccb_scan_ext_t;

typedef struct ccb_translate_ctl {
	ccb_hdr_t	hdr;
	uint32_t	src0_fmt:4;
	uint32_t	src0_sz:5;
	uint32_t	src0_off:3;
	uint32_t	src1_enc:1;
	uint32_t	src1_off:3;
	uint32_t	src1_sz:2;
	uint32_t	output_fmt:2;
	uint32_t	output_sz:2;
	uint32_t	rsvd:1;
	uint32_t	test_val:9;
} ccb_translate_ctl_t;

/*
 * CCB_QUERY_OPCODE_TRANSLATE
 */
typedef struct ccb_translate {
	ccb_translate_ctl_t	control;
	ccb_cmpl_addr_t		completion;
	ccb_byte_addr_t		src0;
	ccb_data_acc_ctl_t	data_acc_ctl;
	ccb_byte_addr_t		src1;
	uint64_t		rsvd;
	ccb_addr_t		dst;
	ccb_tbl_addr_t		vec_addr;
} ccb_translate_t;

typedef struct ccb_select_ctl {
	ccb_hdr_t	hdr;
	uint32_t	src0_fmt:4;
	uint32_t	src0_sz:5;
	uint32_t	src0_off:3;
	uint32_t	rsvd:1;
	uint32_t	src1_off:3;
	uint32_t	rsvd2:2;
	uint32_t	output_fmt:2;
	uint32_t	output_sz:2;
	uint32_t	pad_dir:1;
	uint32_t	rsvd3:9;
} ccb_select_ctl_t;

/*
 * CCB_QUERY_OPCODE_SELECT
 */
typedef struct ccb_select {
	ccb_select_ctl_t	control;
	ccb_cmpl_addr_t		completion;
	ccb_byte_addr_t		src0;
	ccb_data_acc_ctl_t	data_acc_ctl;
	ccb_byte_addr_t		src1;
	uint64_t		rsvd;
	ccb_addr_t		output;
	ccb_tbl_addr_t		tbl;
} ccb_select_t;

#if 0
typedef union ccb {
	ccb_copy_t		copy;
	ccb_imm_data_t		imm_data;
	ccb_fill_t		fill;
	ccb_store_t		store;
	ccb_cas_t		cas;
	ccb_cam_at_t		cam_atomic;
	ccb_at_msg_push_t	atomic_msg_push;
	ccb_sync_nop_t		sync_nop;
	ccb_extract_t		extract;
	ccb_scan_t		scan;
	ccb_scan_ext_t		scan_ext;
	ccb_translate_t		translate;
	ccb_select_t		select;
	uint64_t		dwords[8];
} ccb_t;
#endif

#endif /* _ASM */

#ifdef	__cplusplus
}
#endif

#endif	/* _SYS_CCB_H */
