/*
** Libdax Query
**
** Copyright © 2016 Oracle corp.  All rights reserved.
** The Universal Permissive License (UPL), Version 1.0
**
** Subject to the condition set forth below, permission is hereby granted to any person obtaining a copy of this
** software, associated documentation and/or data (collectively the "Software"), free of charge and under any and
** all copyright rights in the Software, and any and all patent rights owned or freely licensable by each licensor
** hereunder covering either (i) the unmodified Software as contributed to or provided by such licensor, or
** (ii) the Larger Works (as defined below), to deal in both
**
** (a) the Software, and
** (b) any piece of software and/or hardware listed in the lrgrwrks.txt file if one is included with the Software
** (each a “Larger Work” to which the Software is contributed by such licensors),
**
** without restriction, including without limitation the rights to copy, create derivative works of, display,
** perform, and distribute the Software and make, use, sell, offer for sale, import, export, have made, and have
** sold the Software and the Larger Work(s), and to sublicense the foregoing rights on either these or other terms.
**
** This license is subject to the following condition:
** The above copyright notice and either this complete permission notice or at a minimum a reference to the UPL must
** be included in all copies or substantial portions of the Software.
**
** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
** THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
** AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
** CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
** IN THE SOFTWARE.
*/

#ifndef _DAX_QUERY_H
#define	_DAX_QUERY_H

/*
 * libdax_query header file.
 */

#ifdef __cplusplus
extern "C" {
#endif

#include <sys/types.h>
#ifdef __linux
#include <stdint.h>

#ifndef _BOOLEAN_T
#define _BOOLEAN_T
typedef enum { B_FALSE, B_TRUE } boolean_t;
#endif

//typedef  unsigned int uint_t;
#define NANOSEC 1000ULL*1000ULL*1000ULL
#define P2ROUNDUP(SIZE, ALIGN) \
	(((SIZE) + ((ALIGN)-1)) & ~((ALIGN)-1))
#define ISP2(x)         (((x) & ((x) - 1)) == 0)
#define MAX(X,Y) ((X > Y) ? X : Y)
#endif	

/*
 * The DAX requires alignment of output streams and data structures (e.g.,
 * Huffman dictionaries and translate vectors) referenced in CCBs.  No
 * particular alignment is required for input data.
 */
#define	DAX_ALIGNMENT	64


typedef struct dax_enc_ctx {
	uint8_t		*dict_data;	/* kdzk_enc_ctx_dict_data */
	uint32_t	*dict_offsets;	/* kdzk_enc_ctx_dict_offsets */

	uint8_t		*rle_vec;	/* kdzk_enc_ctx_rle_vector */
	uint8_t		rle_vec_width;	/* kdzk_enc_ctx_rle_vector_data_width */
	uint32_t	rle_num_runs;	/* kdzk_enc_ctx_rle_num_of_runs */

	enum dax_enc_flags {
		DAX_NATIVE_TYPE		= 0x001, /* KDZK_NATIVE_TYPE */
		DAX_DATA_ALIGNED	= 0x002, /* KDZK_DATA_ALIGNED */
		DAX_CONSTANT_VALUE	= 0x004, /* KDZK_CONSTANT_VALUE */
		DAX_TRAILING_ZERO	= 0x008, /* KDZK_TRAILING_ZERO */
		DAX_NULL_VALUES		= 0x010, /* KDZK_NULL_VALUES */
		DAX_SIGNED		= 0x020, /* KDZK_SIGNED */
		DAX_PROC_NULLS_WOUT_BV	= 0x040, /* KDZK_PROCESS_NULLS_W/O_BV */
		DAX_DICT		= 0x080, /* KDZK_DICT */
		DAX_RLE			= 0x100, /* KDZK_RLE */
		DAX_DYNP		= 0x200, /* KDZK_DYNP */
		DAX_FIXED		= 0x400, /* KDZK_FIXED */
		DAX_OZIP		= 0x800	 /* KDZK_OZIP */
	} dax_enc_flags_enum_t;		/* kdzk_enc_ctx_flags_enum */

	uint32_t	enc_flags;	/* kdzk_enc_ctx_flags */

	uint8_t		data_width;	/* kdzk_enc_ctx_orig_data_width */
	uint8_t		len_width;	/* kdzk_enc_ctx_len_width */
	uint8_t		flag_width;	/* kdzk_enc_ctx_flag_width */
} dax_enc_ctx_t;

typedef struct dax_vec {
	uint8_t		*data_stream;		/* kdzk_cv_data_stream */
	uint8_t		*length_stream;		/* kdzk_cv_length_stream */
	uint8_t		*flag_stream;		/* kdzk_cv_flag_stream */
	dax_enc_ctx_t	*enc_ctx;		/* kdzk_cv_encoding_context */
	uint64_t	*null_bv;		/* kdzk_cv_null_bit_vector */
	uint64_t	*output_bv;		/* kdzk_cv_output_bit_vector */
	uint32_t 	output_bv_popcnt;	/* kdzk_cv_output_bv_popcnt */
	uint32_t	nrows;			/* kdzk_cv_num_of_rows */
} dax_vec_t;

typedef struct dax_aux {
	uint64_t	*filter;	/* kdzk_aux_filter */
	boolean_t	invert_filter;	/* kdzk_aux_invert_filter */
	uint32_t	current_slot;	/* kdzk_aux_current_slot */
	uint32_t	last_slot;	/* computed from one's vector */
} dax_aux_t;

typedef enum dax_opc {
	DAX_SCAN_NOP = 0,
	DAX_SCAN_EQ,
	DAX_SCAN_NE,
	DAX_SCAN_LT,
	DAX_SCAN_LE,
	DAX_SCAN_GT,
	DAX_SCAN_GE
} dax_scan_op_t;

/* Error codes for dax_*() functions */
typedef enum dax_query_err {
	DAX_QUERY_SUCCESS = 0,	/* Success */
	DAX_QUERY_INTERROR,	/* Internal error */
	DAX_QUERY_LIBERROR,	/* Library error */
	DAX_QUERY_SYSERROR,	/* System error */
	DAX_QUERY_ADIERROR,	/* ADI error */
	DAX_QUERY_DATAERROR,	/* Data error */
	DAX_QUERY_DAXERROR,	/* DAX error */
	DAX_QUERY_NOTSUPPORTED,	/* Not supported */
	DAX_QUERY_THRESHOLD	/* Parameters under threshold for DAX */
} dax_query_err_t;

/*
 * DAX Query API version 1.0
 *
 * RETURN VALUES (for dax_scan, dax_select, dax_translate, and dax_extract)
 *	0  Success
 *     >0  An error occurred; value directly corresponds to dax_query_err_t
 */
typedef struct dax_query_api_1_0 {
	void (*dax_query_set_log_level)(int level);

	int (*dax_scan)(dax_vec_t *res, dax_vec_t *src, dax_scan_op_t op1,
	    dax_vec_t *bound1, dax_scan_op_t op2, dax_vec_t *bound2,
	    dax_aux_t *input);

	int (*dax_select)(dax_vec_t *res, dax_vec_t *src, dax_aux_t *input);

	int (*dax_translate)(dax_vec_t *res, dax_vec_t *src1, dax_vec_t *src2,
	    dax_aux_t *input);

	int (*dax_extract)(dax_vec_t *res, dax_vec_t *src);

	void *(*dax_memset)(void *dest, int c, size_t n);
} dax_query_api_1_0_t;


/*
 * DAX Query API version 1.1
 *
 * RETURN VALUES (for dax_scan, dax_select, dax_translate, and dax_extract)
 *	0  Success
 *     >0  An error occurred; value directly corresponds to dax_query_err_t
 */
typedef struct dax_query_api_1_1 {
	void (*dax_query_set_log_level)(int level);

	int (*dax_scan)(dax_vec_t *res, dax_vec_t *src, dax_scan_op_t op1,
	    dax_vec_t *bound1, dax_scan_op_t op2, dax_vec_t *bound2,
	    dax_aux_t *input);

	int (*dax_select)(dax_vec_t *res, dax_vec_t *src, dax_aux_t *input);

	int (*dax_translate)(dax_vec_t *res, dax_vec_t *src1, dax_vec_t *src2,
	    dax_aux_t *input);

	int (*dax_extract)(dax_vec_t *res, dax_vec_t *src);

	int (*dax_alloc)(size_t length, void **addr);

	int (*dax_free)(void *addr, size_t length);
} dax_query_api_1_1_t;

/*
 * dax_query_init - API version request and library initialization
 *
 * ARGUMENTS
 *
 *   major  Requested major version number
 *   minor  Requested minor version number
 *
 * RETURN VALUES
 *
 *   If the library supports the requested version, dax_query_init() will
 *   return a pointer to that version's API structure; otherwise,
 *   dax_query_init() will return NULL.
 *
 * NOTES
 *
 *   dax_query_init() is to be called once by a process.  After a process
 *   makes a call to dax_query_init(), any thread in that process can call
 *   an API function without calling dax_query_init() again.
 */
void *dax_query_init(uint16_t major, uint16_t minor);

#ifdef __cplusplus
}
#endif

#endif /* _DAX_QUERY_H */
