/*
** Libdax
**
** Copyright © 2016 Oracle corp.  All rights reserved.
** The Universal Permissive License (UPL), Version 1.0
**
** Subject to the condition set forth below, permission is hereby granted to any person obtaining a copy of this
** software, associated documentation and/or data (collectively the "Software"), free of charge and under any and
** all copyright rights in the Software, and any and all patent rights owned or freely licensable by each licensor
** hereunder covering either (i) the unmodified Software as contributed to or provided by such licensor, or
** (ii) the Larger Works (as defined below), to deal in both
**
** (a) the Software, and
** (b) any piece of software and/or hardware listed in the lrgrwrks.txt file if one is included with the Software
** (each a “Larger Work” to which the Software is contributed by such licensors),
**
** without restriction, including without limitation the rights to copy, create derivative works of, display,
** perform, and distribute the Software and make, use, sell, offer for sale, import, export, have made, and have
** sold the Software and the Larger Work(s), and to sublicense the foregoing rights on either these or other terms.
**
** This license is subject to the following condition:
** The above copyright notice and either this complete permission notice or at a minimum a reference to the UPL must
** be included in all copies or substantial portions of the Software.
**
** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
** THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
** AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
** CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
** IN THE SOFTWARE.
*/

#ifndef _SYS_DAX_H
#define	_SYS_DAX_H

#ifdef __KERNEL__
#include "ccb.h"
#else
#include <ccb.h>
#endif

#ifdef	__cplusplus
extern "C" {
#endif

/*
 * DAXIOC_CCB_EXEC dce_hwq_type
 * also used for dax_submit fast trap
 */
#define	DAX_CCB_HWQ_TYPE_MSG	0
#define	DAX_CCB_HWQ_TYPE_QUERY	1

/* DAXIOC_CCB_EXEC dce_ccb_status */
#define	DAX_CCB_STATUS_OK	0
#define	DAX_CCB_STATUS_FAIL	1
#define	DAX_CCB_STATUS_EAGAIN	2
#define	DAX_CCB_STATUS_ENOMAP	3
#define	DAX_CCB_STATUS_EINVAL	4
#define	DAX_CCB_STATUS_ENOACCESS 5
#define	DAX_CCB_STATUS_ENOBUFS	6

/* dax_submit fast trap return values */
#define	DAX_SUBMIT_OK				0
#define	DAX_SUBMIT_ERR_RETRY			1
#define	DAX_SUBMIT_ERR_WOULDBLOCK		2
#define	DAX_SUBMIT_ERR_BUSY			3
#define	DAX_SUBMIT_ERR_THR_INIT			4
#define	DAX_SUBMIT_ERR_ARG_INVAL		5
#define	DAX_SUBMIT_ERR_CCB_INVAL		6
#define	DAX_SUBMIT_ERR_NO_CA_AVAIL		7
#define	DAX_SUBMIT_ERR_CCB_ARR_MMU_MISS		8
#define	DAX_SUBMIT_ERR_NOMAP			9
#define	DAX_SUBMIT_ERR_NOACCESS			10
#define	DAX_SUBMIT_ERR_TOOMANY			11
#define	DAX_SUBMIT_ERR_INTERNAL			12


/*
 * dax_submit flags
 */
#define	DAX_SUBMIT_FLAG_QUERY_CMD	0x0001
#define	DAX_SUBMIT_FLAG_ALL_OR_NONE	0x0002
#ifndef	_ASM

#include <linux/types.h>

#define	DAX_DEV "/dev/dax"
#define DAX_DRIVER_VERSION 2

/*
 * dax device ioctl commands
 */
#define	DAXIOC	'D'

#define	DAXIOC_CCB_THR_INIT	_IOWR(DAXIOC, 1, dax_ccb_thr_init_arg_t)	/* CCB thread initialization */
#define	DAXIOC_CCB_THR_FINI	_IO(DAXIOC,   2)                                /* free CCB thread resources */
#define	DAXIOC_CA_DEQUEUE	_IOWR(DAXIOC, 3, dax_ca_dequeue_arg_t  )	/* CCB CA dequeue */
#define	DAXIOC_CCB_EXEC		_IOWR(DAXIOC, 4, dax_ccb_exec_arg_t    )	/* CCB execution */
#define DAXIOC_VERSION          _IOWR(DAXIOC, 5, long)                          /* get driver version */
#define DAXIOC_DEBUG_VA2PA      _IOWR(DAXIOC, 9, dax_debug_va2pa_t)             /* translate va to pa */

typedef struct dax_debug_va2pa
{
	uint64_t virtual_addr;
	uint64_t physical_addr;
	uint64_t page_size;
} dax_debug_va2pa_t;


/*
 * DAXIOC_CCB_THR_INIT
 */
typedef struct dax_ccb_thr_init_arg {
	uint32_t *dcti_ccb_buf_maxlen;	/* ptr to return uint32_t length */
	uint64_t *dcti_compl_maplen;	/* ptr to return uint64_t mmap length */
	uint64_t *dcti_compl_mapoff;	/* ptr to return uint64_t mmap offset */
} dax_ccb_thr_init_arg_t;

/*
 * DAXIOC_CCB_EXEC
 */
typedef struct dax_ccb_exec_arg {
	uint32_t dce_hwq_type;		/* DAX_CCB_HWQ_TYPE_[MSG|QUERY] */
	uint32_t dce_ccb_buf_len;	/* user buffer length in bytes */
	void     *dce_ccb_buf_addr;	/* user buffer address */
	uint32_t *dce_submitted_ccb_buf_len; /* ptr to return uint32_t len */
	uint64_t *dce_ca_region_off;	/* ptr to return uint64_t byte offset */
	uint32_t *dce_ccb_status;	/* ptr to return uint32_t CCB status */
	uint64_t dce_timer;		/* rob added this */
	uint64_t *dce_nomap_va;		/* ptr to bad virtual address when ret is NOMAP */
} dax_ccb_exec_arg_t;

/*
 * DAXIOC_CA_DEQUEUE
 */
typedef struct dax_ca_dequeue_arg {
	uint32_t dcd_len_requested;	/* byte len of CA to dequeue */
	uint32_t *dcd_len_dequeued;	/* ptr to return uint32_t byte len */
} dax_ca_dequeue_arg_t;

/*
 * fast trap library wrapper in libdax
 */
typedef struct dax_submit_res {
	uint16_t dsr_submitted_ccb_len;	/* byte len of submitted CCBs */
	uint16_t dsr_submitted_ca_off;	/* byte offset of corresponding CAs */
	uint64_t dsr_err_va;		/* valid for ERR_NOMAP, ERR_NOACCESS */
} dax_submit_res_t;

/*
 * Perf Counter defines
 */
#define PERFCOUNT_GET_NODE_COUNT	_IOR('p', 0xB0, void *)
#define PERFCOUNT_DAX_SET_COUNTERS	_IOW('p', 0xBA, void *)
#define PERFCOUNT_DAX_GET_COUNTERS	_IOR('p', 0xBB, void *)
#define PERFCOUNT_DAX_CLEAR_COUNTERS	_IOW('p', 0xBC, void *)

/* The number of DAX engines per node */
#define DAX_PER_NODE		(8)

/* The number of performance counters
 * per DAX engine
 */
#define COUNTERS_PER_DAX	(3)

#ifdef	__cplusplus
}
#endif

#endif /* _ASM */

#endif	/* _SYS_DAX_H */
